/** \brief  'vsi_sudptx_srv' is identical with 'sudp_send2' except for that
 *  this is dedicated for network transfer. File output program used to be
 *  switched by option in 'sudp_send2' though it is obsolate now.
 *  File output function is realized by 'vsi_localwrite_srv' .
 * @Author M.Sekido
 * @Version $Rev: 162 $
 * @Date  $Date: 2015-08-20 15:35:13 +0900 (木, 20  8月 2015) $
 * $Id: vsi_sudptx_srv.cpp 162 2015-08-20 06:35:13Z vlbi $ 
 **/
#include <iostream>
#include <vector>
#include <string>

#include <cstdio>
#include <cstring>
#include <cstdlib>
//#include <boost/algorithm/string.hpp>
using namespace std;

#include "SUDPProto.hpp"
#include "Raw_Proto.hpp"
#include "Mk5B_Frame.hpp"
#include "VDIF_Frame.hpp"
#include "InFile_mdl.hpp"
#include "InMem_mdl.hpp"
#include "Trans_ctl.hpp"
#include "InVSI_mdl.hpp"
#include "StaticVals.hpp"
#include "conf_utl.hpp"

extern "C"{
#include <signal.h>
#include <unistd.h>
}
/** General Configulation data */
Config_set vdif_conf;
char *me;  // My name

void usage(){
  cerr <<  "Usage: " << me 
       << " [-f <filename> | -m <sec> | -d <VSI|VSSP>]  [-p <server port>]\n";
  cerr << " [-R <Hz>] [-t <VDIF | RAW | K5VSSP | MK5B>] [-i <input mode>] [-o <Nout>]\n";
  cerr << " [-w <sec>] [-v <level>] [-x <Normal|Mk5Geo>] [-s <Station Name>:<StnID>]\n";
  //  cerr << " [-O <FILE | NET>] \n";
  cerr << "-----------------------------\n";
  cerr << "  ==DATA SOURCE ==\n";
  cerr << "  -f: sending K5VSSP file with SUDP packets.\n";
    //       << " (if not specified, dummy data will be sent from memory.)\n";
  cerr << "  -m: sending dummy data for duration <sec>.\n";
  cerr << "  -d: Input device type.\n";
  //  cerr << "  -T: Data Frame Type for Input  (default: K5/VSSP32F) \n" ;
  cerr << "  ==Data Output==\n";
  cerr << "  -t: Data Frame Type For Output[VDIF/K5VSSP/MK5B] (default: VDIF) \n" ;
  cerr << "  -o: Number of Output Streams (Default=4) \n" ;
  //   cerr << "  -O: Output Media, [FILE/NET]. (default=NET by sudp) \n" ;
  //  cerr << "  -w: Page Reading Wait time (default: -1 endless) \n" ;
  //  cerr << "  -r: Rate Mbps (default: 100) \n" ;
  cerr << "  ==Control==\n";
  cerr << "  -p: Control Port of this host (default " << DEFAULT_SUDP_PORT
       << ") \n";
  cerr << "  -R: Rate Adjust Frequency (default 100 Hz)\n";
  //  cerr << "  -d: log directory path (default /tmp )\n";
  cerr << " -x: Extraction mode Normal(default) or Mk5Geo(stream 16-19/32 are LSB(14,15ch).\n";

  cerr << "  -i: Input mode in two sorts of formats, ex.)'16Msps/2bit/4ch' or '16:2:4'.\n";
  cerr << "    Default(VSSP): using header info.\n";
  cerr << "    Default(K5VSI): 32Msps/2bit/16ch.\n";
  cerr << "  -s: Station Name in File header must be within 8 chars. (default 'Kashim34').\n";
  cerr << "      Station ID must be 2 character\n";
  cerr << "  -v: verbose level, 0--4 (default -1: no logging).\n";
  cerr << "  ----------- $Id: vsi_sudptx_srv.cpp 162 2015-08-20 06:35:13Z vlbi $ -------" << endl;
  exit(1); 
}

void LogSetup(const int verbose=-1){

  string logdir = "/tmp/";
  string error_file = logdir + me +"_error.txt";
   
  string debug_file = logdir + me + "_dbg.txt";
  Logsystem::debug_logger()->set_stream(debug_file);
  string info_file = logdir + me + "_info.txt";
  Logsystem::info_logger()->set_stream(info_file);
  string warn_file = logdir + me + "_warn.txt";
  Logsystem::warning_logger()->set_stream(warn_file);
  string err_file = logdir + me + "_err.txt";
  Logsystem::error_logger()->set_stream(err_file);
      
  switch(verbose){
  case -1:
    cerr << " No Logging." << endl;
    break;
  case 0:
    Logsystem::debug_logger()->enable_logging(Logger::MINIMUM); 
    Logsystem::info_logger()->enable_logging();
    Logsystem::error_logger()->set_stream(error_file);
    Logsystem::error_logger()->enable_logging();
    break;
  case 1:
    Logsystem::debug_logger()->enable_logging(Logger::SMALL); 
    Logsystem::info_logger()->enable_logging();
    Logsystem::error_logger()->enable_logging();
    Logsystem::warning_logger()->enable_logging();
    break;
  case 2:
    Logsystem::debug_logger()->enable_logging(Logger::MEDIUM); 
    Logsystem::info_logger()->enable_logging();
    Logsystem::error_logger()->enable_logging();
    Logsystem::warning_logger()->enable_logging();
    break;
  case 3:
    Logsystem::debug_logger()->enable_logging(Logger::DETAIL); 
    Logsystem::info_logger()->enable_logging();
    Logsystem::error_logger()->enable_logging();
    Logsystem::warning_logger()->enable_logging();
    break;
  case 4:
  default:
    Logsystem::debug_logger()->enable_logging(Logger::ALL); 
    Logsystem::info_logger()->enable_logging();
    Logsystem::error_logger()->enable_logging();
    Logsystem::warning_logger()->enable_logging();
    break;
  }
  
}


int main(int argc, char *argv[]){
  me = (me = strrchr(argv[0], '/')) == NULL ? argv[0] : me+1; /* My name */ 
  string fname;
  string hostname("localhost");
  int srv_port(DEFAULT_CTRL_PORT);
  bool mflag(false); ///< Flag of memory data
  int mtu(DEFAULT_VDIF_FRAMESIZE +8);  
  int Mbps(100); // Mbps 
  int length(10); // sec
  int Rate_adj_f = 100; // (Hz)

  VLBI_Frame* in_vlbi_frame(0);
  //  DAS_Info das_inf;
  //  int input_wait=-1;
  int verbose_level = -1;
  string device_type;
  DAS_Info indas( 32000000, 16, 2 );
  string inMode;
  bool   mk4geo_mode (false);
  string Stname="Kashim34", Stid="kb";
  int Nout_strm=4; // Number of Output Streams. 
  const int outmedia(0); // Output media, 0=NET, 1=File
  VLBI_Frame::VLBI_Frame_t frame_type(VLBI_Frame::VDIF_frm);

  int k=0;
  while (k >= 0) { 
    switch (k = getopt(argc, argv, "hf:d:p:m:t:w:v:R:i:x:s:o:O:")) { 
    case 'h':   usage();     break;
    case 'f':   fname = optarg; break;
    case 's':   {
      char *cp = strchr(optarg, ':');
      if(cp ==0) Stname = optarg;
      else{
	*cp='\0'; cp++; Stname = optarg; Stid = cp;
      }
      break;

    }
    case 'd':   device_type = optarg; break;
      //    case 's':   mtu = atoi(optarg); break;
    case 't':   
//       if(strncmp(optarg,"VDIF",4)==0){frame_type = VLBI_Frame::VDIF_frm; } 
//       if(strncmp(optarg,"VSSP",4)==0){frame_type = VLBI_Frame::K5VSSP_frm; } 
//       if(strncmp(optarg,"RAW",3)==0){frame_type = VLBI_Frame::K5VSI_frm; } 
//       if(strncmp(optarg,"MK5B",3)==0){frame_type = VLBI_Frame::MK5B_frm; } 

      if(strstr(optarg,"VDIF")!=NULL){frame_type = VLBI_Frame::VDIF_frm; } 
      if(strstr(optarg,"K5VSS")!=NULL){frame_type = VLBI_Frame::K5VSSP_frm; } 
      if(strstr(optarg,"RAW") !=NULL){frame_type = VLBI_Frame::K5VSI_frm; } 
      if(strstr(optarg,"MK5B")!=NULL){frame_type = VLBI_Frame::MK5B_frm; } 
      break;
      //    case 'w':   input_wait = atoi(optarg); break;
    case 'o':   Nout_strm = atoi(optarg); break;
      //    case 'O':    if(strncmp(optarg,"FILE",4)==0) outmedia=1; break;
    case 'i':   inMode = optarg; break;
      //    case 'r':   Mbps = atoi(optarg); break;
    case 'R':   Rate_adj_f = atoi(optarg); break;
    case 'm':   length  = atoi(optarg); mflag = true; break;
      //    case 'a':   hostname = optarg; break;
    case 'p':   srv_port  = atoi(optarg); break;
    case 'x':   if(strncmp(optarg,"Mk5Geo",6) == 0) 
	mk4geo_mode=true; break;
    case 'v':   verbose_level = atoi(optarg); break;
      //    case 'l':   length  = atoi(optarg); break;
    case '?':    usage(); break;
    }
  }

  if(mflag == false && fname.empty() && device_type.empty()){
    cerr << "Error: Input Filename/device/memory is not specified.\n";
    usage();
  }

//   if( out_vlbi_frame == 0 ) out_vlbi_frame = new VDIF_Frame;
//   out_vlbi_frame->Set_StnameID(Stname.c_str(),Stid.c_str());
//   string retname;
//   out_vlbi_frame->Get_Stname(retname);
//   cerr << "Tx Frame Station name=" << retname << endl;

  indas.SetMode(inMode);
//   if( ! inMode.empty() ){
//     int pos = inMode.find("Msps/"); 
//     int sr = atoi(inMode.substr(0,pos).c_str());
//     indas.Set_Srate(sr*1000000);
//     string work = inMode.substr(pos+5);
//     pos = work.find("bit/"); sr = atoi(work.substr(0,pos).c_str());
//     indas.Set_Qbit(sr);
//     work = inMode.substr(pos+4);
//     pos = work.find("ch"); sr = atoi(work.substr(0,pos).c_str());
//     indas.Set_Nch(sr);
//   }
  
  cerr << "Input Mode=" << indas.Mode() << endl;
 
  LogSetup(verbose_level);
  
  VLBI_Proto* vlbip[Nout_strm];

  
//   RAW_Proto rawp[Nout_strm];
//   SUDPProto sudp[Nout_strm];
  VLBI_Frame* out_vlbi_frame[Nout_strm];

  cerr << "Output to Network\n";

  for(int i=0; i<Nout_strm; i++){
    switch( frame_type)
      {
      case VLBI_Frame::K5VSSP_frm:
	out_vlbi_frame[i] = new K5VSSP32_Frame; 
	cerr << "Output Frame: K5VSSP32 Frame\n";
	break;
      case VLBI_Frame::K5VSI_frm: 
	out_vlbi_frame[i] = new K5VSI_Frame; 
	cerr << "Output Frame: K5VSI (Raw)Frame\n";
	break;
      case VLBI_Frame::MK5B_frm:  
	out_vlbi_frame[i] = new Mk5B_Frame;
	cerr << "Output Frame: Mk5B Frame\n"; 
	break;
      case VLBI_Frame::VDIF_frm:
      default:
	out_vlbi_frame[i] = new VDIF_Frame; 
	cerr << "Output Frame: VDIF Frame\n";
	break;
      }
    out_vlbi_frame[i]->Set_StnameID(Stname.c_str(),Stid.c_str());
//     string retname;
//     out_vlbi_frame[i]->Get_Stname(retname);
//     cerr << "Tx Frame Station name=" << retname << endl;

    if(outmedia ==1)  { // Output to file
      vlbip[i] = new RAW_Proto;
      //      cerr << "Output: FILE\n";
    }else {    // Output to Net
      vlbip[i] = new SUDPProto;
      //      cerr << "Output: Network\n";
    }
    
    vlbip[i]->Set_OprMode(TransInfo::TX); // Transmission mode 
    vlbip[i]->Set_Speed(Mbps,true); // Set 100 Mbps
    
    vlbip[i]->Set_Adj_Freq(Rate_adj_f); // Trx-rate adjusting interval
    vlbip[i]->Disp_TxRate(); 
    vlbip[i]->Set_VlbiFrame( out_vlbi_frame[i] ); // Set Output Frame 
    if(mk4geo_mode){
      vlbip[i]->SetExmode_Mk4Geo(); // Extracting LSB 8,9ch->15,16ch
      cerr << "I: Tx mode- Mk4VSI-Geodesy" << endl;
    }
  }
    

  /** Transport Control object */ 
  TransCtl transctl_proto;
  transctl_proto.Set_CtrlPort(srv_port);
  //  out_vlbi_frame->Set_DasInfo(&das_inf); 
  //  in_vlbi_frame->Set_DasInfo(&das_inf); 
  
  //  transctl_proto.Set_DasInfo(&das_inf); // Set DasInfo

  transctl_proto.Set_ApplProto(vlbip, Nout_strm); // Set TransInfo
  //  transctl_proto.Set_ApplProto(vlbip); // Set TransInfo
  
  fprintf(stderr,"Server- Port: %d  MTU= %d \n",  srv_port, mtu);
  fprintf(stderr,"Initial Rate: %d Mbps For %d sec, Rate Adjust by %d cycle\n",
	  Mbps, length, Rate_adj_f);
  /***************************************/

  InDev_Model *input_mdl(0);
  try{
    /** Data Input Device/File */
    if(! fname.empty()){   // sending from file.
      input_mdl = new InFile_Model;
      in_vlbi_frame = new K5VSSP32_Frame;
      in_vlbi_frame->Set_DasInfo( indas );
      input_mdl->Set_VlbiFrame(in_vlbi_frame);
      ((InFile_Model*)input_mdl)->Set_infile(fname.c_str());  
      fprintf(stderr,"Data source : File(%s)  \n", fname.c_str()); 
    }else if(!device_type.empty()){ // Data input from Device
      //    if( strncmp(device_type.c_str(),"VSI",3) == 0) 
      input_mdl = new K5_VSI[Nout_strm];
      in_vlbi_frame = new K5VSI_Frame[Nout_strm];
      for(int i=0; i<Nout_strm; i++){
	((K5VSI_Frame*)in_vlbi_frame +i)->Set_DasInfo( indas );
	((K5_VSI*)input_mdl+i)->Set_VlbiFrame( ((K5VSI_Frame*)in_vlbi_frame)+i );
      }
      //if( strncmp(device_type.c_str(),"VSSP",4) == 0) input_mdl = new K5_VSSP;
      fprintf(stderr,"Data source : Device(%s)  \n", device_type.c_str()); 
      if( input_mdl->TestDev() < 0){
	throw "Failed to open K5/VSI device file";
      }
    }else{ // data from memory
      input_mdl = new InMem_Model(length);
      //      input_mdl->start(); // Memory update start
      in_vlbi_frame = new K5VSI_Frame;
      in_vlbi_frame->Set_DasInfo( indas );
      input_mdl->Set_VlbiFrame(in_vlbi_frame);
      fprintf(stderr,"Data source: Memory   \n") ;
    }
    //  input_mdl->Set_PageWait(input_wait);
    input_mdl->Set_StnameID(Stname.c_str(),Stid.c_str());
    if(!device_type.empty()){ // Data input from K5VSI Device
      for(int i=1; i<Nout_strm; i++){
	((K5_VSI*)input_mdl +i)->Set_StnameID(Stname.c_str(),Stid.c_str());
      }
    }

    transctl_proto.Set_InputDev(input_mdl); // Set Input device 
    if(! transctl_proto.SetupCheck()) exit(1);
    //  input_mdl->Ready();
    //  input_mdl->start(); // reading data Started 

    //  signal(SIGINT,SIG_IGN);

    transctl_proto.start(); // Transaction server starting
  
  
    cerr << "TransCtrl Thread ...";
    transctl_proto.join();
    cerr << "joined\n";
    }// End of try

  catch(const char* msg){
    cerr << "Error:" << msg << endl;
  }  
  for(int i=0; i<Nout_strm; i++)  vlbip[i]->Disp_RateOff();  

  /** Memory Release */
  if( input_mdl != 0 ){
    if(!device_type.empty()) // Data input from K5VSI Device
      delete [] input_mdl;
    else  delete input_mdl;
  }
  if( in_vlbi_frame != 0 ) delete in_vlbi_frame;
  for(int i=0; i<Nout_strm; i++)
    if( out_vlbi_frame[i] != 0 ){ 
      delete out_vlbi_frame[i]; 
      out_vlbi_frame[i]=0;
    }
  /****************************/

  if(Logsystem::Test_Logsys(Logsystem::DEBUG)) 
    Logsystem::debug_logger()->close_stream();
  if(Logsystem::Test_Logsys(Logsystem::INFO))   
    Logsystem::info_logger()->close_stream();
}


