/** 
 * @Author M.Sekido
 * @Version $Rev: 56 $
 * @Date  $Date: 2013-07-03 20:40:14 +0900 (水, 03  7月 2013) $
 * $Id: sudp_recv3.cpp 56 2013-07-03 11:40:14Z vlbi $ 
 */
#include <iostream>
#include <string>
#include <vector>
using namespace std;
#include <boost/algorithm/string.hpp>

#include <time.h>
//#include <sys/types.h>
//#include <unistd.h>
#include <cstdio>
#include <cstring>
#include <cstdlib>

#include "VDIF_Frame.hpp"
#include "VTProto.hpp"
#include "Trans_ctl.hpp"
#include "Raw_Proto.hpp"
#include "InNet_mdl.hpp"
#include "Dumper.hpp"
#include "StaticVals.hpp"
#include "LogSet.hpp"
#include "conf_utl.hpp"

/** General Configulation data */
Config_set vdif_conf;

extern "C"{
#include <unistd.h>
#include <sys/types.h>
}

char *me;  // My name

void usage(){
  cerr << "Usage: " << me <<" [-p <ctrl port>] [-P <data port>] [-g <n>]\n";
  cerr << "    [-m <mode>] [-i <info logfile>] [-x <Norm|Mk5Geo>] [-a <ackhost>] [-b <ackport>] \n"; // add miya 2015/6/23 [-a <ackhost>] 2016/2/3 [-b <ackport>]
  cerr << "    [-s <Stname>:<ID>] [-v <n>]\n";
  //  cerr << "   [-m <out-mode>] [-M <in-mode>] [-l <sec>]\n";
  //  cerr << "  [-r <Mbps>]  [-d <logging file>]  [-a <host>]\n";
  cerr << "  ---<Control>-----------------------------------------------\n";
  cerr << "  Server mode data receiver of VDIF/SDUP stream\n";
  cerr << " -p <port>, port number for communication with client. default=9903.\n";
  cerr << "  ---<Input>-----------------------------------------------\n";
  cerr << " -P <port>, port number for receiving data stream. default=9901.\n";
  cerr << " -m: Input data mode. ex.: 32Msps/2bit/16ch is default.\n";
  cerr << "  ---<Output>-----------------------------------------------\n";
  cerr << " -g: Number of max output threads.\n";
  cerr << "    Supporting Output format: VDIF, VSSP32. \n";
  cerr << " -x: Extraction mode Normal(default) or Mk5Geo(stream 16-19/32 are \n";
  cerr << "     LSB(14,15ch).\n";
  cerr << "  ---<Information>------------------------------------------\n";
  cerr << " -a <ackhost>: AckHost (default:no AckPacket).\n"; // add miya 2015/6/23 [-a <ackhost>] 
  cerr << " -b <ackport>, AckData sending port number. default=29000\n"; // add miya 2016/2/3 [-b <ackport>]
  cerr << " -s <Stname>:<ID>: Station name(8 byte) and ID(2byte) for K5/VSSP32\n";
  cerr << "     file header information. \n";
  //cerr << " -i <info file>:  Logging information(packet-loss etc.).\n";
  cerr << " -v: verbose level, 0--4 (default -1: no logging).\n";
  cerr << " -l: <log file name>.\n";
  cerr << " -n: Setting frame size by number of divisions/sec.\n"; // add 2014/5/28 miya
  cerr << "$Id: sudp_recv3.cpp 56 2013-07-03 11:40:14Z vlbi $\n";
  cerr << "-------------------------------------------------------" << endl;
  exit(1);
  //  cerr << " -o <file>[:<file2>..] Output to files for multiple out.\n";
//   cerr << " -t <FMT>: Writing format, VDIF | VSSP. (default:VSSP32).\n";
//   cerr << " -n <nch>, Number of channels (1|4) for output. default=4.\n";
//   cerr << " -g <group>[:<grp>...], Writing Group number of nch. 0,1,2..\n";
//   cerr << " -m <'xxMsps/xbit/xxch'>: Output Data mode\n";
//   cerr << " -M <'xxMsps/xbit/xxch'>: Input  Data mode\n";
//   cerr <<"        Msps -> [04Msps|08Msps|16Msps|32Msps|64Msps]\n";
//   cerr << "        bit  -> [1bit|2bit|4bit|8bit] \n";
//   cerr << "        ch   -> [16ch|08ch|04ch|02ch|01ch] \n";
//   cerr << " -m <mode> Recording mode. ex) 32Msps/2bit/16ch, Default:16Msps/1bit/4ch.\n";
//  cerr << " -s <byte> Packet size (default=1500).\n";
  //  cerr << " -r <Mbps> Expected rate (default=100).\n";
  //cerr << " -l <sec> length of receiving data(default=0 means endless).\n";
}


int main(int argc, char *argv[]){
  me = (me = strrchr(argv[0], '/')) == NULL ? argv[0] : me+1; /* My name */ 

  char logstr[256];
  int loglevel(0);
  int ndiv=10; // Number of frames in 1sec in VDIF output file
  //  for (int i = 0; i < argc; i++) { printf("%d = %s\n", i, argv[i]);  }
  int rxport(DEFAULT_DATA_PORT);
  int ctrlport(DEFAULT_CTRL_PORT2);
  int ackport(ACK_DEFAULT_DATA_PORT);  // add miya 2016/2/3
  //  string ip("0.0.0.0");
  string def_outmode = "2048Msps/1bit/1ch";
  //LogSet logset(me,-1);
  int ngrp(1) ;// Max number of output groups
  string inmode = "2048Msps/1bit/1ch";
  string outmode = "2048Msps/1bit/1ch";
  bool mk4geo_mode = false;
  bool log_l = false;
  string Stname("Kashim34"), Stid("kb");
  string hostname; // add miya 2015/6/23
	
  int k=0;
	//while (k >= 0) { 
    //switch (k = getopt(argc, argv, "h:p:P:v:g:m:i:x:s:")) {
  while(( k = getopt(argc, argv, "h:p:P:v:g:m:x:a:s:l:n:b:")) != -1){  // n: add 2014/5/28 miya  add miya 2015/6/23 a: 2016/2/3 b:
    switch (k){
    case 'l':
      {
	Time_dat t1; t1.Set_time_to_now();
	string ydhms = t1.Print_time(13);
	sprintf(logstr,"%s%s_%s%d",me,ydhms.c_str(),optarg,getpid()); 
	
      }
      break;
    case 'p':   ctrlport  = atoi(optarg); break;
    case 'P':   rxport  = atoi(optarg); break;
    case 'm':   inmode = optarg; break;
      //case 'i':   logset.Set_infofile(optarg); break;
      //case 'v':   logset.Set_level(atoi(optarg)); break;
    case 'v':   loglevel = atoi(optarg); log_l = true; break;
    case 'g':    ngrp = atoi(optarg); break;
    case 's':   {
      char *cp = strchr(optarg,':');
      if( cp != NULL){ *cp = '\0'; cp++; Stid = cp;}
      Stname =optarg;
      break;}
    case 'x':   if(strncmp(optarg,"Mk5Geo",6) == 0) mk4geo_mode=true; break;
    case 'a':   hostname  = optarg; break;  // add miya 2015/6/23
    case 'b':   ackport  = atoi(optarg); break; // add miya 2016/2/3
    case 'n':   ndiv = atoi(optarg); break;  // add 2014/5/28 miya
    case 'h':    usage(); break;
    case '?':    cerr << "Missing following parameters\n"; usage(); break;
    case ':':    cerr << "Invalid option\n"; usage(); break;
    }
  }

  LogSet logset(logstr,-1);

  if (log_l){ logset.Set_level(loglevel); }
  logset.Start();
  
  DAS_Info   outdas; /// commentout miya
  DAS_Info   indas;
  /// comment miya start------------------------------------------
  ///if( indas.SetMode(inmode)){
  ///  cerr << "Initial Input mode=" << indas.Mode() << endl;
  ///}else{
  ///  cerr << "Error at mode setting.\n";
  ///  usage();
  ///};
  /// comment miya end---------------------------------------------
  
  /// commentout miya start------------------------------------------
   if(! outdas.SetMode(outmode) || ! indas.SetMode(inmode)){
     cerr << "Error at mode setting\n";
     usage();
   }
  /// comment miya end---------------------------------------------
   
  VDIF_Frame in_frame;
  in_frame.Set_DasInfo(indas); 

//  cerr << "Intial DAS = " << outdas.Mode() << endl;
  VTProto inp;
  inp.Set_VlbiFrame(&in_frame);
  inp.Set_DataPort(rxport);
  //  inp.Set_TransLength(length); // sec
  //  inp.Set_Speed(rate); // (Mbps)
  inp.Disp_RxRate();
  inp.Set_Ack_ONOFF(false); // add miya 2015/6/23
  if( ! hostname.empty() ) { // add miya 2015/6/23
	  inp.Set_AckHost(hostname.c_str());  // add miya 2015/6/23
	  inp.Set_AckSend_ONOFF(true); // add miya 2015/6/23
  	  inp.Set_AckDataPort(ackport); // add miya 2016/2/3
  }else{inp.Set_AckSend_ONOFF(false);} // add miya 2015/6/23
  cerr << "Innet Intializing..." << flush;
  InNet_Model innet;
  cerr << " Set Protcol to Innet..." << flush;
  innet.Set_protocol(&inp);
  cerr << " Frame Set..." << flush;
  innet.Set_VlbiFrame(&in_frame);
  cerr << " DasInfo Set..." << flush;
  innet.Set_DasInfo(indas);
  cerr << endl;

  /** add 2014/6/27 miya*/
  if( innet.Ready() < 0 ){
	cerr << "Error in innet.Ready()\n";
	exit(1);
  }
  /** Receiving start in advance. 2014.6.26 by M.S.*/
  innet.start();

  /** TxRxing Control object */ 
  TransCtl transctl_proto;
  transctl_proto.Set_CtrlPort(ctrlport);

  VLBI_Proto* outp[ngrp];   // Raw output protocol

  //K5VSSP32_Frame out_frame[ngrp];   // VSSP32 output Frame 
  VDIF_Frame out_frame[ngrp];   
	
  for(int i=0; i<ngrp; i++){
    outp[i] = new RAW_Proto;
	out_frame[i].Set_DasInfo(outdas); /// add miya
	outp[i]->Set_VlbiFrame(&out_frame[i]); /// add miya
	out_frame[i].Set_StnameID(Stname.c_str(), Stid.c_str());
  	out_frame[i].Set_Nframe_sec( ndiv ); /// add 2014/5/28 miya
  	cerr << "vdif Div=" << out_frame[i].Nframe_sec() << endl;
    if(mk4geo_mode){
      outp[i]->SetExmode_Mk4Geo(); // Extracting LSB 8,9ch->15,16ch
      cerr << "I: Rcving mode["<< i << "] Mk4VSI-Geodesy" << endl;
    }
  }
	cerr << "Innet mae..." << flush;
  transctl_proto.Set_ApplProto(outp, ngrp); // Set TransInfo
  //  transctl_proto.Set_ApplProto(&outp[0], ngrp); // Set TransInfo
  transctl_proto.Set_InputDev(&innet);
  if(! transctl_proto.SetupCheck()) exit(1);

	cerr << "Innet ato..." << flush;
  transctl_proto.start(); // Transaction server starting
	
//   while(1){
//     Logsystem::info_logger()->Tstamp()
//       << transctl_proto.DumperState() << "\n";
//     Logsystem::info_logger()->Flush();
//     sleep(1);
//   }

  transctl_proto.join();  // waiting for ending

  inp.Disp_RateOff();  

  for(int i=0; i < ngrp; i++){ delete outp[i]; }
  logset.Stop();
}
