/** 
 * @Version $Rev: 143 $
 * @Author M.Sekido
 * @Date  $Date: 2014-10-17 09:32:28 +0900 (金, 17 10月 2014) $
 * $Id: utlty.cpp 143 2014-10-17 00:32:28Z vlbi $ 
 */
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "maxmin_v2.hpp"
#include "utlty.hpp"

long line_num(const char *file)
{
  FILE *fp;
  long int ln;
  const char *wc_cmd = "/usr/bin/wc -l ";
  char command[256];
  
  sprintf(command, "%s %s", wc_cmd, file);
  
  if( (fp = popen(command, "r")) == NULL){
    fprintf(stderr, "Error: Cannot open pipe for file %s \n",file );
    exit(EXIT_FAILURE);
  }
  int ret = fscanf(fp, "%ld", &ln);
  pclose(fp);
  if(ret < 1) return ret;
  
  return(ln);
}


/**
 * Compute GCM via euclid 
 */
long  euclid_gcm(const long  a, const long  b)
{
  //  double r, a1,b1;
  long  r, a1,b1;
  

  b1=a;
  r = b;
  
  while (r != 0.0){ 
    //    r = fmod(a1,b1);
    a1 = b1;
    b1 = r;
    r = a1 % b1;
  }
   
  return(b1);
}

/// Return GCM of array
long  arr_gcm(const double x[], int ndata)
{
 long imin;
 int	i;
 
 Maxmin_db <double> maxmin_db;
 
 maxmin_db.Max_min1(x,ndata);
 if( maxmin_db.Max() > 2.e9){
   cerr << "Error ing arr_gcm(): data(Max=" << maxmin_db.Max() 
	<< ") is too large >2.e9\n";
   cerr << "     Data must be less than 1.e9\n";
   exit(1);
 }
 imin = (long)rint(maxmin_db.Min());
 if (imin ==0){
   fprintf(stderr, "Error:Minimum of array in arr_gcm is 0\n");
   exit(1);
 }
 
 for(i=0; i<ndata; i++){
   long ix = (long)rint(x[i]);
   //   if (fmod( ix, imin) != 0.0)
     imin = euclid_gcm( ix, imin); 
 }
 
 return (imin);
}


double arr_mean(double x[], int ndata){
  double mx=0;
  int i;

  for(i=0; i<ndata; i++) mx += x[i];
  return(mx/ndata);
}  

int arr_log10f(float *xarr, float *logarr, int n){
  int i;

  for(i=0; i<n; i++){
    if( xarr[i] < 0.){ 
      fprintf(stderr, "Error: cannot take log10(xarr[%d] = %f)\n",
	      i, xarr[i]);
      return(-i);
    }
    if( xarr[i] == 0.)
      logarr[i] = -45; // Minimum of Float value
    else 
      logarr[i] = log10f( xarr[i]);
  }
  return(1);
  
}









